"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""

from j40_model.simulation_heuristic import simulation
from j40_model.optimization_model import optimization


def interventions_main(df_e, el_sets,
                       energy_burden_target, cost_table, opt=False, **kwargs):

    if opt:
        res_bdg, res_tr = optimization(df_e, el_sets,
                                       energy_burden_target, cost_table, **kwargs)
    else:
        res_bdg, res_tr = simulation(df_e, el_sets,
                                     energy_burden_target, cost_table)

    # get tract results per building and join
    per_bdg = df_e[['fips']].copy()
    for c_gen in ['cs', 'cw']:
        el_per_tract = df_e[['fips', 'weight']].loc[
            el_sets[f'k_{c_gen}']].groupby('fips').sum()
        for lbl in [x for x in res_tr.keys() if c_gen in x]:
            tr_map = res_tr[lbl].div(el_per_tract.weight,
                                     axis=0).fillna(0).to_dict()
            per_bdg[lbl] = per_bdg.fips.map(tr_map)
    per_bdg = res_bdg.join(per_bdg.drop('fips', axis=1))

    return per_bdg, res_tr



"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""

col_map_base = {
                'state': 'Sate Initials',
                'fips': 'Tract #',
                'weight': '# of Households',
                'tenure': 'Tenure',
                'year_built': 'Year Built',
                'ami': 'AMI Level (%)',
                'baseline_ec': 'Baseline Annual Energy Bill ($)',
                'baseline_eb': 'Baseline Energy Burden'
                }

col_map_results = {'d_wth': '# Buildings Weatherized',
                   'C_wth': 'Weatherization Investments ($)',
                   'c_wth': 'Weatherization Annualized Investments ($)',
                   'd_rts': 'Rooftop Solar Deployment (kW)',
                   'C_rts': 'Rooftop Solar Investments ($)',
                   'c_rts': 'Rooftop Solar Annualized Investments ($)',
                   'g_rts': 'Generation Rooftop Solar (kWh)',
                   'g_cs': 'Generation Community Solar (kWh)',
                   'g_cw': 'Generation Community Wind (kWh)',
                   'ec': 'Energy Bill After Intervention ($)',
                   'eb': 'Energy Burden After Intervention',
                   'd_cs': 'Community Solar Deployment (kW)',
                   'C_cs': 'Community Solar Investments ($)',
                   'c_cs': 'Community Solar Annualized Investments ($)',
                   'd_cw': 'Community Wind Deployment (kW)',
                   'C_cw': 'Community Wind Investments ($)',
                   'c_cw': 'Community Wind Annualized Investments ($)',
                   'solar_kw_potential': 'Rooftop Solar Potential (kW)',
                   'wind_kw_potential': 'Community Wind Potential (kW)',
                   'jobs_per_home': 'Jobs per Home Weatherized',
                   }
col_map_results.update(col_map_base)



"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""

from j40_model.aux_functions import data_corrections, aggregate_per_tract, eb_percentage
from j40_model.interventions import interventions_main
import pandas as pd
from j40_model.nomenclature import col_map_results, col_map_base
import numpy as np


class CommunityCase:

    def __init__(self, df, cost_table, eb_target):

        df = data_corrections(df)
        df['baseline_ec'] = df[['annual_elec_cost_dol_per_year',
                                'annual_gas_cost_dol_per_year',
                                'annual_fuel_cost_dol_per_year']].sum(axis=1)
        df['baseline_eb'] = df['baseline_ec'] / df['annual_income']

        self.community_table = df

        self.cost_table = cost_table

        self.el_ami = ['80-100%']
        self.el_tenure = ['OWNER', 'RENTER']
        self.fips = df.fips.unique()
        self.policies = ['weatherization', 'rooftop_solar',
                         'community_solar', 'community_wind']

        self.eb_target = eb_target
        self.results_per_bdg = pd.DataFrame()
        self.results_per_tract = pd.DataFrame()

    def run_intervention(self, optimization=False, **kwargs):
        df = self.community_table

        df_e = df.loc[df['baseline_eb'] > self.eb_target]

        eligible = {}
        eligible_index = []
        for p in self.policies:
            el_df = df_e.loc[(df_e.ami.isin(self.el_ami))
                             & (df_e.tenure.isin(self.el_tenure))]
            eligible.update({p: el_df})
            eligible_index += list(el_df.index)

        df_e = df_e.loc[np.unique(eligible_index)].copy()

        # eligibility sets of k
        e_w = eligible['weatherization']
        el_sets = {
            'k_rts': eligible['rooftop_solar'].index.to_list(),  # eligible for rooftop solar
            'k_cs': eligible['community_solar'].index.to_list(),  # eligible for community solar
            'k_cw': eligible['community_wind'].index.to_list(),  # eligible for community wind
            'k_w_el': e_w.loc[e_w.heating_fuel.isin(
                ['ELECTRICITY'])].index.to_list(),  # weatherization + electricity
            'k_w_gs': e_w.loc[e_w.heating_fuel.isin(
                ['BOTTLED GAS', 'UTILITY GAS'])].index.to_list(),  # weatherization + gas
            'k_w_of': e_w.loc[e_w.heating_fuel.isin(
                ['WOOD', 'COAL', 'FUEL OIL', 'OTHER'])].index.to_list(),  # weatherization + fuel
        }

        # run intervention
        self.results_per_bdg, self.results_per_tract = interventions_main(
            df_e=df_e, el_sets=el_sets, energy_burden_target=self.eb_target,
            cost_table=self.cost_table, opt=optimization, **kwargs)

    def _eligible_df(self, just_eligible=False):
        if just_eligible:
            df = self.results_per_bdg.join(self.community_table)
        else:
            df = self.community_table.join(self.results_per_bdg).fillna(0)
            df.loc[~df.index.isin(self.results_per_bdg.index), 'ec'] = df.loc[
                ~df.index.isin(self.results_per_bdg.index), 'baseline_ec']
            df.loc[~df.index.isin(self.results_per_bdg.index), 'eb'] = df.loc[
                ~df.index.isin(self.results_per_bdg.index), 'baseline_eb']

        return df

    def get_results_per_building(self, just_eligible=False):
        df = self._eligible_df(just_eligible=just_eligible)
        df = df[col_map_results.keys()]
        df = df.rename(columns=col_map_results)
        df = eb_percentage(df)
        return df

    def get_results_per_tract(self, just_eligible=False):
        df = self._eligible_df(just_eligible=just_eligible)
        # aggregate building results per tract and join
        summed_keys = ['d_wth', 'C_wth', 'c_wth',
                       'd_rts', 'C_rts', 'c_rts', 'g_rts',
                       'g_cs',  'g_cw',
                       'solar_kw_potential', 'wind_kw_potential'
                       ]
        weighted_keys = ['baseline_ec', 'baseline_eb', 'ec', 'eb', 'jobs_per_home']
        t_df = aggregate_per_tract(df, summed_keys, weighted_keys)
        t_df = t_df.join(self.results_per_tract).fillna(0)
        t_df = t_df.rename(columns=col_map_results)
        t_df = eb_percentage(t_df)
        return t_df

    def get_base_per_tract(self):
        df = self.community_table
        summed_keys = []
        weighted_keys = ['baseline_ec', 'baseline_eb']
        t_df = aggregate_per_tract(df, summed_keys, weighted_keys)
        t_df = t_df.rename(columns=col_map_base)
        t_df = eb_percentage(t_df)
        return t_df



"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""

def data_corrections(df):
    """
    This method contains ad-hoc corrections to the community file
    ideally those corrections should be integrated in new versions of the community file
    """
    df['weight'].fillna(0, inplace=True)
    df = df[(df['annual_elec_cost_dol_per_year'] + df['annual_gas_cost_dol_per_year']
             + df['annual_fuel_cost_dol_per_year']) < 0.5 * df['annual_income']].copy()
    df.annual_wind_cf *= 8760

    # income target eligibility
    df['baseline_ec'] = df[['annual_elec_cost_dol_per_year', 'annual_gas_cost_dol_per_year',
                            'annual_fuel_cost_dol_per_year']].sum(axis=1)
    df['baseline_eb'] = df['baseline_ec'] / df['annual_income']

    return df


def aggregate_per_tract(df, summed_keys, weighted_keys):
    all_keys = summed_keys + weighted_keys

    w_df = df[['fips', 'weight']].copy()
    for k in all_keys:
        w_df[k] = df[k] * df.weight
    tw_df = w_df.groupby('fips').sum()

    for k in weighted_keys:
        tw_df[k] = tw_df[k] / tw_df.weight

    return tw_df[all_keys + ['weight']]


def eb_percentage(df):
    for col in [c for c in df.columns if 'burden' in c.lower()]:
        df[col +' (%)'] = df[col] * 100.0
    return df



"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""

import pandas as pd
from pyomo.environ import *
from pyomo.opt import SolverFactory


def total_cost_cost(c, r, lt):

    return c / (r / (1 - (1 + r) ** -lt))


def optimization(df, eligible, eb_target, cost_table, **kwargs):

    if kwargs:
        inv_rate = kwargs['inv_rate']
        eb_weight = kwargs['eb_weight']
        budget_limit = kwargs['budget_limit']
        solver = kwargs['solver']

    else:
        inv_rate = 0.03
        eb_weight = 1
        budget_limit = 1e10
        solver = 'cplex'

    m = ConcreteModel()

    print('writing model', solver)
    # sets
    m.k_rts = Set(initialize=eligible['k_rts'])
    m.k_cs = Set(initialize=eligible['k_cs'])
    m.k_cw = Set(initialize=eligible['k_cw'])

    m.k_w_el = Set(initialize=eligible['k_w_el'])
    m.k_w_gs = Set(initialize=eligible['k_w_gs'])
    m.k_w_of = Set(initialize=eligible['k_w_of'])
    m.k_w = m.k_w_el | m.k_w_gs | m.k_w_of

    m.k = Set(initialize=df.index.to_list())
    m.T = Set(initialize=df.loc[m.k]['fips'].unique())
    m.Tk_cs = Set(m.T, initialize=df.loc[m.k_cs].reset_index()[
        ['fips', 'index']].groupby('fips')['index'].apply(list).to_dict())
    m.Tk_cw = Set(m.T, initialize=df.loc[m.k_cw].reset_index()[
        ['fips', 'index']].groupby('fips')['index'].apply(list).to_dict())

    m.Tk_cg = Set(m.T, initialize=df.loc[m.k_cs | m.k_cw].reset_index()[
        ['fips', 'index']].groupby('fips')['index'].apply(list).to_dict())

    # params
    m.WS = Param(m.k_w, initialize=df.loc[m.k_w]['weatherization_savings_per_home'].to_dict())
    m.rts_prod = Param(m.k_rts, initialize=df.loc[m.k_rts]['annual_solar_cf'].to_dict())
    m.cs_prod = Param(m.k_cs, initialize=df.loc[m.k_cs]['annual_solar_cf'].to_dict())
    m.wnd_prod = Param(m.k_cw, initialize=(df.loc[m.k_cw]['annual_wind_cf']).to_dict())

    m.cw_max_kw = Param(m.k_cw, initialize=(df.loc[m.k_cw]['wind_kw_potential']).to_dict())
    m.rts_max_kw = Param(m.k_rts, initialize=df.loc[m.k_rts]['solar_kw_potential'].to_dict())

    m.n_bdg = Param(m.k, initialize=df.loc[m.k]['weight'].to_dict())
    m.income = Param(m.k, initialize=df.loc[m.k]['annual_income'].to_dict())

    m.C_rts = Param(initialize=cost_table.loc['rooftop_solar']['cost'])
    m.C_cs = Param(initialize=cost_table.loc['community_solar']['cost'])
    m.C_cw = Param(initialize=cost_table.loc['community_wind']['cost'])
    m.C_wth = Param(m.k_w, initialize=(df.loc[m.k_w]['weatherization_cost_per_home'] *
                                       cost_table.loc['weatherization']['cost']).to_dict())

    m.Lt_rts = Param(initialize=cost_table.loc['rooftop_solar']['lifetime'])
    m.Lt_cs = Param(initialize=cost_table.loc['community_solar']['lifetime'])
    m.Lt_cw = Param(initialize=cost_table.loc['community_wind']['lifetime'])
    m.Lt_wth = Param(initialize=cost_table.loc['weatherization']['lifetime'])

    m.C_el = Param(m.k, initialize=df.loc[m.k]['annual_elec_cost_dol_per_year'].to_dict())
    m.C_gs = Param(m.k, initialize=df.loc[m.k]['annual_gas_cost_dol_per_year'].to_dict())
    m.C_of = Param(m.k, initialize=df.loc[m.k]['annual_fuel_cost_dol_per_year'].to_dict())

    m.P_el = Param(m.k, initialize=df.loc[m.k]['avg_elec_price_dol_per_kwh'].to_dict())
    m.R = Param(initialize=inv_rate)

    m.kcs_t = Param(m.k_cs, initialize=df.loc[m.k_cs]['fips'].to_dict())
    m.kcw_t = Param(m.k_cw, initialize=df.loc[m.k_cw]['fips'].to_dict())

    # vars
    m.d_wth = Var(m.k_w, domain=NonNegativeReals)  # number of buildings weatherized
    m.d_rts = Var(m.k_rts, domain=NonNegativeReals)  # MW of solar per building deployed
    m.d_cs = Var(m.T, domain=NonNegativeReals)  # MW deployed in the track
    m.d_cw = Var(m.T, domain=NonNegativeReals)  # MW deployed in the track

    m.g_rts = Var(m.k, domain=NonNegativeReals)  # MWh rooftop generation per building
    m.g_cs = Var(m.k, domain=NonNegativeReals)  # MWh community solar generation
    m.g_cw = Var(m.k, domain=NonNegativeReals)  # MWh community wind generation

    m.c_rts = Var(m.k_rts, domain=NonNegativeReals)  # Annualized costs of solar per building
    m.c_cs = Var(m.T, domain=NonNegativeReals)  # Annualized costs of community solar per tract
    m.c_cw = Var(m.T, domain=NonNegativeReals)  # Annualized costs of community wind per tract
    m.c_wth = Var(m.k_w, domain=NonNegativeReals)  # Annualized costs of weatherization per building
    m.nel_tract = Var(m.T, domain=Reals)  # Net-load tract

    m.eld = Var(m.k, domain=NonNegativeReals)  # electricity demand per building kwh
    m.nel = Var(m.k, domain=Reals)  # net electricity demand per building kWh
    m.ec = Var(m.k, domain=NonNegativeReals)  # energy consumption per building kwh
    m.eg = Var(m.k, domain=NonNegativeReals)  # energy generation per building kwh
    m.eb = Var(m.k, domain=NonNegativeReals)  # energy burden per building
    m.eb_p_plus = Var(m.k, domain=NonNegativeReals)  # burden gap
    m.eb_p_minus = Var(m.k, domain=NonNegativeReals)  # burden over_target

    m.ci = Var(domain=NonNegativeReals)  # annual investment costs
    m.bp_function = Var(domain=NonNegativeReals)  # burden penalty
    m.budget_limit = Param(initialize=budget_limit)
    m.eb_weight = Param(initialize=eb_weight, mutable=True)  # always initialized as 1

    def const_7(m, k):
        return m.c_wth[k] == m.d_wth[k] * m.C_wth[k] * (m.R / (1 - (1 + m.R) ** -m.Lt_wth))

    m.const_7 = Constraint(m.k_w, rule=const_7)

    def const_8(m, k):
        return m.g_rts[k] == m.d_rts[k] * m.rts_prod[k]

    m.const_8 = Constraint(m.k_rts, rule=const_8)

    def const_8_2(m, k):
        return m.g_rts[k] == 0

    m.const_8_2 = Constraint(m.k - m.k_rts, rule=const_8_2)

    def const_9(m, k):
        t = m.kcs_t[k]
        return m.g_cs[k] == m.d_cs[t] * m.cs_prod[k] / sum(m.n_bdg[kcs] for kcs in m.Tk_cs[t])

    m.const_9 = Constraint(m.k_cs, rule=const_9)

    def const_9_2(m, k):
        return m.g_cs[k] == 0

    m.const_9_2 = Constraint(m.k - m.k_cs, rule=const_9_2)

    def const_10(m, k):
        t = m.kcw_t[k]
        return m.g_cw[k] == m.d_cw[t] * m.wnd_prod[k] / sum(m.n_bdg[kcw] for kcw in m.Tk_cw[t])

    m.const_10 = Constraint(m.k_cw, rule=const_10)

    def const_10_2(m, k):
        return m.g_cw[k] == 0

    m.const_10_2 = Constraint(m.k - m.k_cw, rule=const_10_2)

    def const_11(m, k):
        return m.c_rts[k] == m.d_rts[k] * m.C_rts * (m.R / (1 - (1 + m.R) ** -m.Lt_rts))

    m.const_11 = Constraint(m.k_rts, rule=const_11)

    def const_12(m, t):
        return m.c_cs[t] == m.d_cs[t] * m.C_cs * (m.R / (1 - (1 + m.R) ** -m.Lt_cs))

    m.const_12 = Constraint(m.T, rule=const_12)

    def const_13(m, t):
        return m.c_cw[t] == m.d_cw[t] * m.C_cw * (m.R / (1 - (1 + m.R) ** -m.Lt_cw))

    m.const_13 = Constraint(m.T, rule=const_13)

    def const_14(m, k):
        return m.eld[k] == m.C_el[k] / m.P_el[k] - m.d_wth[k] * m.WS[k] * m.C_el[k] / m.P_el[k]

    m.const_14 = Constraint(m.k_w_el, rule=const_14)

    def const_15(m, k):
        return m.eld[k] == m.C_el[k] / m.P_el[k]

    m.const_15 = Constraint(m.k - m.k_w_el, rule=const_15)

    def const_16(m, k):
        return m.ec[k] == m.eld[k] * m.P_el[k] + m.C_gs[k] + m.C_of[k]

    m.const_16 = Constraint(m.k - m.k_w_gs - m.k_w_of, rule=const_16)

    def const_17(m, k):
        return m.ec[k] == m.eld[k] * m.P_el[k] + m.C_gs[k] - (m.C_gs[k] * m.d_wth[k] * m.WS[k]) + m.C_of[k]

    m.const_17 = Constraint(m.k_w_gs, rule=const_17)

    def const_18(m, k):
        return m.ec[k] == m.eld[k] * m.P_el[k] + m.C_gs[k] + m.C_of[k] - (m.C_of[k] * m.d_wth[k] * m.WS[k])

    m.const_18 = Constraint(m.k_w_of, rule=const_18)

    def const_19(m, k):
        return m.eg[k] == m.g_rts[k] + m.g_cs[k] + m.g_cw[k]

    m.const_19 = Constraint(m.k, rule=const_19)

    def const_20(m, k):
        return m.eb[k] == (m.ec[k] - m.eg[k] * m.P_el[k]) / m.income[k]

    m.const_20 = Constraint(m.k, rule=const_20)

    def const_21(m, k):
        return m.eb[k] - eb_target == m.eb_p_plus[k] - m.eb_p_minus[k]

    m.const_21 = Constraint(m.k, rule=const_21)

    def const_22(m, k):
        pass  # Defined in the variable construction

    def const_23(m, k):
        return m.d_wth[k] <= 1

    m.const_23 = Constraint(m.k_w, rule=const_23)

    def const_27(m, k):
        return m.eld[k] >= m.eg[k]

    m.const_27 = Constraint(m.k, rule=const_27)

    def const_28(m, k):
        return m.d_rts[k] * m.n_bdg[k] <= m.rts_max_kw[k]

    m.const_28 = Constraint(m.k_rts, rule=const_28)

    def const_29(m, t):
        return m.d_cw[t] <= sum(m.cw_max_kw[k] for k in m.Tk_cw[t])

    m.const_29 = Constraint(m.T, rule=const_29)

    def objective_first_row(m):
        return m.ci == sum(m.c_cs[t] + m.c_cw[t] for t in m.T) + sum(m.c_rts[k] * m.n_bdg[k] for k in m.k_rts) \
               + sum(m.c_wth[k] * m.n_bdg[k] for k in m.k_w)

    m.objective_first_row = Constraint(rule=objective_first_row)

    def objective_second_row(m):
        return m.bp_function == sum(100 * m.eb_p_plus[k] * m.n_bdg[k] for k in m.k)

    m.objective_second_row = Constraint(rule=objective_second_row)

    def budget_constraint(m):
        return (m.ci - m.eb_weight * m.budget_limit) / 1e6 <= 0

    m.budget_limit_constraint = Constraint(rule=budget_constraint)

    # solving to parameterize the function:

    # 1 - minimize equity
    m.ObjectiveFunction = Objective(rule=m.bp_function)
    opt = SolverFactory(solver)
    results = opt.solve(m)

    # 2 - minimize budget with constrained equity
    equity = m.bp_function.value
    def equity_constraint(m):
        return m.bp_function <= equity
    m.equity_constraint = Constraint(rule=equity_constraint)
    m.del_component(m.ObjectiveFunction)
    m.ObjectiveFunction = Objective(rule=m.ci)
    opt = SolverFactory(solver)
    results = opt.solve(m, tee=True)

    res_df = pd.DataFrame(index=df.index)
    for v in [m.d_wth, m.c_wth,  # weatherization variables
              m.d_rts, m.c_rts,  # rooftop_variables
              m.g_rts, m.g_cs, m.g_cw,  # generation variables
              m.eg, m.ec, m.eld, m.nel,  # energy_cost_variables
              m.eb, m.eb_p_plus, m.eb_p_minus,  # energy burden variables
              ]:
        res_df = res_df.join(pd.Series(v.get_values(), name=v.getname()))

    res_df['C_wth'] = total_cost_cost(res_df['c_wth'], inv_rate, m.Lt_wth.value)
    res_df['C_rts'] = total_cost_cost(res_df['c_rts'], inv_rate, m.Lt_rts.value)

    df_tract = pd.DataFrame(index=df.fips.unique())
    df_tract = df_tract.sort_index()
    for vt in [m.d_cs, m.c_cs,
               m.d_cw, m.c_cw]:
        df_tract = df_tract.join(pd.Series(vt.get_values(), name=vt.getname()))

    df_tract['C_cs'] = total_cost_cost(df_tract['c_cs'], inv_rate, m.Lt_cs.value)
    df_tract['C_cw'] = total_cost_cost(df_tract['c_cw'], inv_rate, m.Lt_cw.value)

    return res_df, df_tract



# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='j40model',
    version='1.0.0',
    description='Base simulation and optimization modules for J40 tool',
    long_description=readme,
    author='Miguel Heleno',
    author_email='miguelhelenoa@lbl.gov',
    include_package_data=True,
    packages=['j40_model'],
)



"""
*** Copyright Notice ***

Justice 40 Tool (J40 Tool) Copyright (c) 2023, The Regents of the University
of California, through Lawrence Berkeley National Laboratory (subject to receipt ofany required approvals from the U.S.
Dept. of Energy). All rights reserved.

If you have questions about your rights to use or distribute this software,
please contact Berkeley Lab's Intellectual Property Office at
IPO@lbl.gov.

NOTICE.  This Software was developed under funding from the U.S. Department
of Energy and the U.S. Government consequently retains certain rights.  As
such, the U.S. Government has been granted for itself and others acting on
its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
Software to reproduce, distribute copies to the public, prepare derivative
works, and perform publicly and display publicly, and to permit others to do so.
"""


import pandas as pd


def annual_cost(c, r, lt):

    return c * (r / (1 - (1 + r) ** -lt))


def simulation(df, eligible, eb_target,
               cost_table, inv_rate=0.03):

    df = df.copy()
    df_tract = pd.DataFrame(index=df.fips.unique())

    for k in eligible.keys():
        df[k] = 0
        df.loc[eligible[k], k] = 1

    #  weatherization
    df['eld'] = df['annual_elec_cost_dol_per_year'] / df['avg_elec_price_dol_per_kwh'] * (
            1 - df.k_w_el * df.weatherization_savings_per_home)

    df['ec'] = (df.eld * df.avg_elec_price_dol_per_kwh
                + df.annual_gas_cost_dol_per_year * (1 - df.weatherization_savings_per_home * df.k_w_gs)
                + df.annual_fuel_cost_dol_per_year * (1 - df.weatherization_savings_per_home * df.k_w_of))

    df['d_wth'] = df[['k_w_el', 'k_w_gs', 'k_w_of']].sum(axis=1)
    df['C_wth'] = df.d_wth * df.weatherization_cost_per_home * cost_table.loc['weatherization']['cost']
    df['c_wth'] = annual_cost(df.C_wth, inv_rate, cost_table.loc['weatherization']['lifetime'])

    # Calculate DER needs under net-metering conditions based on energy burden
    g = ((df.ec - eb_target * df.annual_income
          ) / df.avg_elec_price_dol_per_kwh).clip(lower=0)
    g = pd.concat([g, df.eld], axis=1).min(axis=1)

    # rts deployment is limited by eligibility and potential
    df['d_rts'] = pd.concat([df.k_rts * g / df.annual_solar_cf,
                             df.solar_kw_potential / df.weight], axis=1).min(axis=1)
    df['g_rts'] = df.d_rts * df.solar_kw_potential
    df['C_rts'] = df['d_rts'] * cost_table.loc['rooftop_solar']['cost']
    df['c_rts'] = annual_cost(df['C_rts'], inv_rate, cost_table.loc['rooftop_solar']['lifetime'])

    # update DER generation needs and net_load
    g = g - df.d_rts * df.g_rts
    net_d = df.eld - df.g_rts

    # Community Solar generation
    # get the max generation per building based on the building with
    # minimum available net load in the tract
    tract = pd.concat([df[['fips', 'k_cs']], net_d, g], axis=1)
    net_g_tract = tract.loc[tract.k_cs == 1].groupby('fips')[0].min()
    g_tract = tract.loc[tract.k_cs == 1].groupby('fips')[1].max()

    df['g_cs'] = df.fips.map(pd.concat([net_g_tract,
                                        g_tract], axis=1).min(axis=1).to_dict()) * df.k_cs

    df_tract['d_cs'] = pd.concat([df.fips, df.g_cs * df.weight / df.annual_solar_cf],
                                 axis=1).groupby('fips').sum()

    df_tract['C_cs'] = df_tract.d_cs * cost_table.loc['community_solar']['cost']
    df_tract['c_cs'] = annual_cost(df_tract.C_cs, inv_rate,
                                   cost_table.loc['community_solar']['lifetime'])

    # update DER generation needs and net_load
    net_d = net_d - df.g_cs

    # Community Wind - repeat the same as for com solar
    # only deploys CW if k_cs =! k_cw
    tract = pd.concat([df[['fips', 'k_cw']], net_d, g], axis=1)
    net_g_tract = tract.loc[tract.k_cw == 1].groupby('fips')[0].min()
    g_tract = tract.loc[tract.k_cw == 1].groupby('fips')[1].max()

    df['g_cw'] = df.fips.map(pd.concat([net_g_tract,
                                        g_tract], axis=1).min(axis=1).to_dict()) * df.k_cw

    df_tract['d_cw'] = pd.concat([df.fips, df.g_cw * df.weight / df.annual_wind_cf],
                                 axis=1).groupby('fips').sum()
    df_tract['C_cw'] = df_tract.d_cw * cost_table.loc['community_wind']['cost']
    df_tract['c_cw'] = annual_cost(df_tract.C_cw, inv_rate,
                                   cost_table.loc['community_wind']['lifetime'])

    df['eg'] = df[['g_rts', 'g_cs', 'g_cw']].sum(axis=1)
    df['eb'] = df['ec']/df.annual_income
    res_bdg = df[['d_wth', 'C_wth', 'c_wth', 'd_rts', 'C_rts',
                  'c_rts', 'g_rts', 'g_cs', 'g_cw', 'ec', 'eld', 'eg', 'eb']]

    return res_bdg, df_tract
